/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/
/**********************************************************************
 FILE     : %M%
 PURPOSE  : MPEG motion estimator using Ultra SPARC V8/9 VIS intructions
 NOTES    : The motion estimation can be accelerated by 2-3 times if
          : procedure call overhead is reduced, i.e by changing the
          : function in a way that the motion estimation of the full
          : image is calculated in a single procedure call.
 AUTHOR   : Niels Mache
 COPYRIGHT: 1998 Niels Mache, Dirk Farin
 DATE     : August 1998
 VERSION  : %I%  %G%
 
 Niels Mache             email: mache@informatik.uni-stuttgart.de
 IPVR, Inst. of Parallel and Distributed High Performance Systems
 D 70565 Stuttgart, GERMANY.    voice: +49 711.7816.254, fax: 248
 http://www.informatik.uni-stuttgart.de/ipvr/bv/personen/mache

 **********************************************************************/


#include <stdlib.h>
#include "vis_types.h"
#include "vis_proto.h"

#include "coding/bcVIS.hh"

void BlockComparator_VIS::SetImageBitmap(const Bitmap<Pixel>& bm,bool frame,bool top)
{
  if (frame)
    d_image = bm.AskFrame_const();
  else
    d_image = bm.AskField_const(top);
}


bool BlockComparator_VIS::Error(BCData* bcdata,
                                     const ComparePos& cmppos,
                                     uint32& error,
                                     uint32  maxerror,
                                     int subfact) const
{
  int refx = cmppos.ref_x;
  int refy = cmppos.ref_y;

  const BCData_Bitmaps* bcdat = static_cast<BCData_Bitmaps*>(bcdata);
  assert(bcdat);

  const MVPixel*const* ref = bcdat->AskPixels(refx,refy);

  vis_u8  *ip, *rp;
  vis_d64 *sp1, *sp2;                     /* 8-byte aligned start point in 
                                             frame1 and frame2 */
  vis_d64 sd1, sd11;
  vis_d64 sd2, sd22;
  vis_d64 s11, s10;
  vis_d64 dist;                          /* accumulated sum of differences */
  union {vis_d64 d64;
         unsigned long long ull;} result;
  int dx  = cmppos.img_x;
  int endy = cmppos.img_y + bcdat->d_blk_height;

#if defined (VIS_SUM64)
  unsigned long long maxsum = maxerror;
#endif

  // check block width and subsampling
  // code is written for subsampling distance 1
  assert(bcdat->d_blk_width == 16);
  assert(subfact == 1);

#if (VIS_GSR)
  // Use GSR segment register to speed up unaligned memory access.
  // This can code accelerate unaligned memory access by approx. 4-8 times
  // but need additional setup time.
  // Use this code for succesive (multiple) 16x16 searches.
  ip = &d_image    [cmppos.img_y][0];  // Zeiger ins Originalbild
  sp1 = (vis_d64 *) vis_alignaddr(ip, dx);
  int sp1_gsr=vis_read_gsr();
  int sp1_step=(&d_image[subfact][0] - &d_image[0][0]) >> 3;

  rp = &ref        [refy][0];  // Zeiger ins Referenzbild
  sp2 = (vis_d64 *) vis_alignaddr(rp, refx);
  int sp2_gsr=vis_read_gsr();
  int sp2_step=(&ref[subfact][0] - &ref[0][0]) >> 3;
#endif

  // zero distance count
  dist = vis_fzero();
  for (int iy=cmppos.img_y , ry=refy ;
       iy<endy ;
       iy+=subfact,ry+=subfact)
    {
#if (!VIS_GSR)
      ip = (vis_u8*) &d_image[iy][0];  // Zeiger ins Originalbild
      rp = (vis_u8*) &ref[ry][0];      // Zeiger ins Referenzbild
#endif

#if (!VIS_GSR)
      /* load 8 bytes of source data and align */
      sp1 = (vis_d64 *) vis_alignaddr(ip, dx);
#else
      // load source data segmented by scale register (GSR)
      vis_write_gsr(sp1_gsr);
#endif

      // feed superscalar execution units
      s10 = sp1[0];
      s11 = sp1[1];
      sd1 = vis_faligndata(s10, s11);
      // swap n' align
      s10 = sp1[2];
      sd11 = vis_faligndata(s11, s10);
#if (VIS_GSR)
      sp1 += sp1_step;
#endif

#if (!VIS_GSR)
      /* load 8 bytes of reference data and align */
      sp2 = (vis_d64 *) vis_alignaddr(rp, refx);
#else
      // load reference data segmented by scale register (GSR)
      vis_write_gsr(sp2_gsr);      
#endif

      // feed superscalar execution units
      s10 = sp2[0];
      s11 = sp2[1];
      sd2 = vis_faligndata(s10, s11);
      // swap n' align
      s10 = sp2[2];
      sd22 = vis_faligndata(s11, s10);
#if (VIS_GSR)
      sp2 += sp2_step;
#endif

      // feed superscalar execution units, trap dependencies
      dist = vis_pdist(sd1, sd2, dist);
      dist = vis_pdist(sd11, sd22, dist);

      result.d64 = dist;
#if !defined (VIS_SUM64)
      error = result.ull;
#endif

#if 1
# if defined (VIS_SUM64)
      if (result.ull >= maxsum) return false;
# else
      if (error >= maxerror) return false;
# endif
#endif
    }

#if defined (VIS_SUM64)
   if (subfact != 1)  error = (uint32)result.ull * (subfact*subfact); 
#else
  if (subfact != 1)  error *= subfact*subfact;
#endif

  return true;
}
