/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CODING_MOTIONEST_HH
#define CODING_MOTIONEST_HH

#include "types.hh"
#include "image/bitmap.hh"
#include <stdlib.h>


struct SearchParam
{
  // Suchbereich
  int minx,maxx;
  int miny,maxy;

  // Halfpelsuche ?
  bool halfpel;

  // Frame-Suche ?
  bool framesearch;
  
  // Field-Search: TopField ?
  bool top_field;

  // Suchblockgroesse
  int blk_width,blk_height;

  // Referenzbitmap (enthaelt ein Frame oder beide Fields)
  const Bitmap<Pixel>* bitmap;
};




struct MEData
{
  virtual ~MEData() { }
};


class MotionEstimator
{
public:
  MotionEstimator() { }
  virtual ~MotionEstimator() { }


  virtual MEData* CreateMEData(const SearchParam&,const MEData* reuse=NULL) = 0;
  virtual bool    CanReuse(const MEData*) const { return false; }

  virtual void SetSearchParams(MEData*, const Bitmap<Pixel>* bitmap, bool frame, bool top=true) = 0;
  virtual bool Estimate(int x,int y,     // linke obere Ecke des Blocks im Bild
                        int& mx,int& my, // Rueckgabe des berechneten Vektors
                        double& error,   // Rueckgabe des Fehlers
                        double  upper_error_bound=512.0) = 0;   // Maximaler Fehler

protected:
  void AdjustRange(const SearchParam&,   // maximaler Suchbereich
                   int x,int y,          // linke obere Ecke des Blocks
                   int width,int height, // Bildgroesse
                   int& left,int& right,int& top,int& bottom) const;  // Rueckgabe des Suchbereichs
};




struct BCData
{
  virtual ~BCData() { }
};


class BlockComparator
{
public:
  BlockComparator() { }
  virtual ~BlockComparator() { }

  static BlockComparator* CreateDefaultBC();

  virtual BCData* CreateBCData  (const SearchParam&) = 0;
  virtual void    SetImageBitmap(const Bitmap<Pixel>&,bool frame,bool top=true) = 0;

  struct ComparePos
  {
    int img_x,img_y;  // Linke obere Ecke im der Originalbild-Bitmap, bei der die Suche beginnen soll.
    int ref_x,ref_y;  // Linke obere Ecke in der Referenz-Bitmap, bei der die Suche beginnen soll.
  };


  // Mindestens eine der folgenden Methoden "Error" und "ErrorMulti" muss umdefiniert werden !!!
  // TODO: brauchen wir den Rueckgabewert ?

  virtual bool Error     (BCData*,
                          const ComparePos&,  // Positionen der 16x16 Bloecke und Half-Pel-Flag
                          uint32& error,      // Rueckgabe des Gesamtfehlers.
                          uint32  maxerror,   // Obere Grenze fuer den Fehler.
                          // Bei Ueberschreitung kann abgebrochen werden.
                          int subfact=1       // Unterabtastungsfaktor
                          ) const;
  virtual void ErrorMulti(BCData*,
                          int nVecs,
                          const ComparePos*,  // Positionen der 16x16 Bloecke und Half-Pel-Flag
                          uint32* errors,     // Rueckgabe des Gesamtfehlers.
                          uint32  maxerror,   // Obere Grenze fuer den Fehler.
                          // Bei Ueberschreitung kann abgebrochen werden.
                          int subfact=1       // Unterabtastungsfaktor
                          ) const;
};




#define MV_COMMON_PART \
  /* Suchbereich ggf. verkleinern, wenn wir uns zu nahe am Rand befinden. */ \
 \
  int left,right,top,bottom; \
  AdjustRange(c_medata->sparam, x,y, c_img_width, c_img_height, left,right,top,bottom); \
 \
  BlockComparator::ComparePos cmppos; \
  cmppos.img_x = x; \
  cmppos.img_y = y; \
\
  /* maximal erlaubter Gesamtfehler */ \
  uint32 max_mv_error = (uint32)(upper_error_bound * c_pixels_per_block); \
\
  uint32 nomv_error = 0;  /* Fehler, falls der alte (uebergebene) Motionvektor benutzt wird. */ \
  bool   nomv_vector_valid = false; /* Ob, der alte Motionvektor benutzt werden darf. */ \
\
  int halfpelshiftfact = (c_medata->sparam.halfpel ? 1 : 0); \
\
  /* MAE-Fehler berechnen, fuer den Fall, dass der MV nicht veraendert wird. */ \
\
  if (mx>=left && mx<=right && my>=top && my<=bottom) \
    { \
      cmppos.ref_x = (x<<halfpelshiftfact)+mx; \
      cmppos.ref_y = (y<<halfpelshiftfact)+my; \
\
      /* Fehler mit altem MV berechnen. */ \
\
      if (d_blkcmp->Error(c_bcdata,cmppos,nomv_error,max_mv_error)) \
	{ \
	  double mean_nomv_error = ((double)nomv_error)/c_pixels_per_block; \
\
	  /* Wenn der Fehler sehr gering ist, dann gleich diesen Vektor nehmen und nicht weitersuchen. */ \
          if (mean_nomv_error <= 4.0) /* Empfohlener Wert: 1 */ \
	    {\
	      return true;\
	    }\
          else /* sonst die Schwelle fuer den zu suchenden MV herabsetzen. */ \
            if (mean_nomv_error <= 3.0) /* Empfohlener Wert: 3 */ \
	      {\
		max_mv_error = nomv_error/2;\
              }\
	    else\
	      {\
                max_mv_error = nomv_error*29/32; /* *0.9 */ \
              }\
	  nomv_vector_valid = true; \
	}\
      else\
	{ assert(0); } \
    }\
  /* MVs durchsuchen */ \
  int x0 = (x<<halfpelshiftfact); /* Zentrumskoordinaten (im Referenzbild) */ \
  int y0 = (y<<halfpelshiftfact);

#endif
