/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/motionest_1dfs.hh"
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include "global.hh"

#define LOG_OUTPUT 0

struct MEData_1DFS : public MEData
{
  ~MEData_1DFS() { delete bcdata; }

  BCData*     bcdata;
  SearchParam sparam;
};


MotionEstimator_1DFS::MotionEstimator_1DFS()
{
  d_blkcmp = BlockComparator::CreateDefaultBC();
}

MotionEstimator_1DFS::~MotionEstimator_1DFS()
{
  delete d_blkcmp;
}


MEData* MotionEstimator_1DFS::CreateMEData(const SearchParam& spar,const MEData*)
{
  MEData_1DFS* medata = new MEData_1DFS;
  medata->bcdata = d_blkcmp->CreateBCData(spar);
  medata->sparam = spar;

  return medata;
}

void MotionEstimator_1DFS::SetSearchParams(MEData* medata, const Bitmap<Pixel>* bitmap,
                                           bool frame, bool top)
{
  MEData_1DFS* med = dynamic_cast<MEData_1DFS*>(medata);
  assert(med);

  c_medata = med;
  c_bcdata = med->bcdata;
  c_frame_flag = frame;
  c_top_flag   = top;
  c_img_width  = bitmap->AskWidth();
  c_img_height = bitmap->AskHeight();
  c_pixels_per_block = med->sparam.blk_width * med->sparam.blk_height;

  d_blkcmp->SetImageBitmap(*bitmap,frame,top);
}


bool MotionEstimator_1DFS::Estimate(int x,int y, int& mx,int& my,
					   double& error,double upper_error_bound)
{
  MV_COMMON_PART
    /*
    cout << "OTS x0: " << x0 << " y0: " << y0 << endl;
  cout << "left: " << left << " right: " << right << " top: " << top << " bottom: " << bottom << endl;
    */

  uint32 maxerror = (uint32)(upper_error_bound*c_pixels_per_block);
  uint32 err;

  int stepsize = (c_medata->sparam.halfpel ? 2 : 1);

  // volle horizontale Suche

  uint32 minerror=MASK31;
  int bestx;
  for (int x=left;x<=right;x++)
    {
      uint32 err;
      cmppos.ref_y = y0;
      cmppos.ref_x = x0+x;
      d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);

      if (err<minerror)
        { minerror=err; bestx=x; }
    }


  // volle vertikale Suche

  int besty=0;
  for (int y=top;y<=bottom;y++)
    {
      uint32 err;
      cmppos.ref_y = y0+y;
      cmppos.ref_x = x0+bestx;
      d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);

      if (err<minerror)
        { minerror=err; besty=y; }
    }

  mx = bestx;
  my = besty;


  if (mx != 0 || my != 0)
    {
      int xrange = (right-left+1)/2;
      int yrange = (bottom-top+1)/2;
      int left2   = bestx-xrange; if (left2  <left  ) left2  =left;
      int right2  = bestx+xrange; if (right2 >right ) right2 =right;
      int top2    = besty-yrange; if (top2   <top   ) top2   =top;
      int bottom2 = besty+yrange; if (bottom2>bottom) bottom2=bottom;

      for (int x=left2;x<=right2;x++)
        {
          uint32 err;
          cmppos.ref_y = y0+besty;
          cmppos.ref_x = x0+x;
          d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
          
          if (err<minerror)
            { minerror=err; bestx=x; }
        }
      
      for (int y=top2;y<=bottom2;y++)
        {
          uint32 err;
          cmppos.ref_y = y0+y;
          cmppos.ref_x = x0+bestx;
          d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
          
          if (err<minerror)
            { minerror=err; besty=y; }
        }

      mx = bestx;
      my = besty;
    }

  // Half-Pels drumrum testen

  if (c_medata->sparam.halfpel &&
      mx-1 >= left && mx+1 <= right && my-1 >= top && my+1 <= bottom)
    {
      uint32 err;
      cmppos.ref_x=x0+mx-1; cmppos.ref_y=y0+my-1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx-1; my=my-1; }
      cmppos.ref_x=x0+mx  ; cmppos.ref_y=y0+my-1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx  ; my=my-1; }
      cmppos.ref_x=x0+mx+1; cmppos.ref_y=y0+my-1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx+1; my=my-1; }
      cmppos.ref_x=x0+mx-1; cmppos.ref_y=y0+my  ; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx-1; my=my  ; }
      cmppos.ref_x=x0+mx+1; cmppos.ref_y=y0+my  ; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx+1; my=my  ; }
      cmppos.ref_x=x0+mx-1; cmppos.ref_y=y0+my+1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx-1; my=my+1; }
      cmppos.ref_x=x0+mx  ; cmppos.ref_y=y0+my+1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx  ; my=my+1; }
      cmppos.ref_x=x0+mx+1; cmppos.ref_y=y0+my+1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=mx+1; my=my+1; }
    }

  //  cout << "MV: " << mx << " / " << my << endl;

  return true;
}
