/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/motionest_ntss.hh"
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include "global.hh"

struct MEData_NTSS : public MEData
{
  ~MEData_NTSS() { delete bcdata; }

  BCData*     bcdata;
  SearchParam sparam;
};


MotionEstimator_NTSS::MotionEstimator_NTSS()
{
  d_blkcmp = BlockComparator::CreateDefaultBC();
}

MotionEstimator_NTSS::~MotionEstimator_NTSS()
{
  if (d_blkcmp) delete d_blkcmp;
}


MEData* MotionEstimator_NTSS::CreateMEData(const SearchParam& spar,const MEData*)
{
  MEData_NTSS* medata = new MEData_NTSS;
  medata->bcdata = d_blkcmp->CreateBCData(spar);
  medata->sparam = spar;

  return medata;
}

void MotionEstimator_NTSS::SetSearchParams(MEData* medata, const Bitmap<Pixel>* bitmap,
                                                  bool frame, bool top)
{
  MEData_NTSS* med = dynamic_cast<MEData_NTSS*>(medata);
  assert(med);

  c_medata = med;
  c_bcdata = med->bcdata;
  c_frame_flag = frame;
  c_top_flag   = top;
  c_img_width  = bitmap->AskWidth();
  c_img_height = bitmap->AskHeight();

  c_pixels_per_block = med->sparam.blk_width * med->sparam.blk_height;

  d_blkcmp->SetImageBitmap(*bitmap,frame,top);
}

bool MotionEstimator_NTSS::Estimate(int x,int y, int& mx,int& my,
					   double& error,double upper_error_bound)
{
  MV_COMMON_PART

  // MVs durchsuchen (logarithmische Suche)

  int hstep = max(-left,right )/2;   // aktuelle horizontale Schrittweite
  int vstep = max(-top ,bottom)/2;   // aktuelle vertikale   Schrittweite
  int motx=0,moty=0;  // aktuelles Zentrum der Suche
  bool first=true;    // Nur fuer in erstem Schritt der MV-Suche ==true;

  uint32 err[3][3];   // Fehler der 9 Suchpositionen. Zentrum liegt in [1][1].

  int    minx=1,miny=1; // Beste Suchposition, minx,miny aus 0,1,2
                        // (Initialisierung nicht notwendig, beruhigt aber den Compiler)
  uint32 minval=MASK31; // Fehler bei der besten Suchposition.

  int stepsize = (c_medata->sparam.halfpel ? 2 : 1);

  // Schrittweite der Log.-Suche bis Schrittweite 1 verfeinern.
  while (hstep>0 || vstep>0)
    {
      uint32 e[3][3];

      if (first)
        {
          for (int dx=-1;dx<=1;dx++)
            for (int dy=-1;dy<=1;dy++)
              {
                if (dx>=left && dx<=right &&
                    dy>=top  && dy<=bottom)
                  {
                    cmppos.ref_x = dx + x0;
                    cmppos.ref_y = dy + y0;
                    d_blkcmp->Error(c_bcdata,cmppos,e[dx+1][dy+1],MASK31);
                  }
                else
                  e[dx+1][dy+1]=MASK31;
              }
        }


      // 8 Positionen um das Zentrum (+ Zentrum beim ersten Mal) testen.
      for (int dx=-1;dx<=1;dx++)
	for (int dy=-1;dy<=1;dy++)
	  {
	    if (!first && dx==0 && dy==0) continue;  // Zentrum nur beim ersten mal testen.

	    int sx = motx + dx*hstep;     // zu testender Motion-Vektor ist (sx;sy)
	    int sy = moty + dy*vstep;

	    // Nur testen, wenn MV innerhalb der erlaubten Grenzen.
	    if (sx>=left && sx<=right &&
		sy>=top  && sy<=bottom)
	      {
		// Position des Blocks im Referenzbild
		cmppos.ref_x = sx + x0;
		cmppos.ref_y = sy + y0;

		// Fehler berechnen
		if (d_blkcmp->Error(c_bcdata,cmppos,err[dx+1][dy+1],minval /*,subsampling*/))
		  if (err[dx+1][dy+1]<minval)
		    {
		      // Fehler ist besser als alles bisherige -> Position und Fehler merken
		      minx=dx+1; miny=dy+1;
		      minval=err[dx+1][dy+1];
		    }
	      }
	  }



      if (first)
        {
          uint32 min_inner_8=MASK31;
          uint32 min_outer_8=MASK31;
          uint32 center=e[1][1];
          int mininx=0,mininy=0;

          for (int dx=0;dx<=2;dx++)
            for (int dy=0;dy<=2;dy++)
              {
                if (dx!=1 || dy!=1)
                  {
                    if (e  [dx][dy] < min_inner_8) { min_inner_8 = e  [dx][dy]; mininx=dx; mininy=dy; }
                    if (err[dx][dy] < min_outer_8)   min_outer_8 = err[dx][dy];
                  }
              }

          if (center < min_inner_8 && center < min_outer_8)
            {
              mx=my=0;
              error=((double)center)/c_pixels_per_block;
              return true;
            }

          if (min_inner_8 <= min_outer_8)
            {
              {for (int dx=-1;dx<=1;dx++)
                for (int dy=-1;dy<=1;dy++)
                  {
                    if (dx+mininx>=left && dx+mininx<=right &&
                        dy+mininy>=top  && dy+mininy<=bottom)
                      {
                        cmppos.ref_x = mininx+dx + x0;
                        cmppos.ref_y = mininy+dy + y0;
                        d_blkcmp->Error(c_bcdata,cmppos,e[dx+1][dy+1],MASK31);
                      }
                    else
                      e[dx+1][dy+1]=MASK31;
                  }}

              uint32 mininner=MASK31;
              {for (int dx=0;dx<=2;dx++)
                for (int dy=0;dy<=2;dy++)
                  {
                    if (e[dx][dy] < mininner) { mininner = e[dx][dy]; mx=mininx+dx-1; my=mininy+dy-1; }
                  }}
              error = ((double)mininner)/c_pixels_per_block;
              return true;
            }
        }

      // Beste Position fuer naechsten Suchschritt ins Zentrum verschieben

      // Zentrum der Suche verschieben
      motx += (minx-1)*hstep;
      moty += (miny-1)*vstep;

      err[1][1] = err[minx][miny]; // Fehler ist ja schon bekannt.
      minx=1;
      miny=1;

      // Suchschrittweite halbieren
      hstep /= 2;
      vstep /= 2;
    }

  /* Falls neuer Vektor nicht oder nur unwesentlich (s.o.) besser ist, als alter MV,
     den alten MV weiterbenutzen. */
  if (nomv_vector_valid && err[1][1] > max_mv_error)
    {
      error = ((double)nomv_error)/c_pixels_per_block;
      return true;
    }

  /* Falls aber der neue Vektor besser ist, den neuen verwenden. */
  if (err[1][1] <= max_mv_error)
    {
      // MV zurueckgeben
      mx = motx;
      my = moty;

      error = ((double)err[1][1])/c_pixels_per_block;
      return true;
    }

  cout << upper_error_bound << endl;
  cout << err[1][1] << " > " << max_mv_error << endl;

  // Sonst konnte leider kein MV mit der angegebenen oberen Fahlerschranke gefunden werden.
  return false;
}
