/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/motionest_ots.hh"
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include "global.hh"

#define LOG_OUTPUT 0

struct MEData_OTS : public MEData
{
  ~MEData_OTS() { delete bcdata; }

  BCData*     bcdata;
  SearchParam sparam;
};


MotionEstimator_OTS::MotionEstimator_OTS()
{
  d_blkcmp = BlockComparator::CreateDefaultBC();
}

MotionEstimator_OTS::~MotionEstimator_OTS()
{
  delete d_blkcmp;
}


MEData* MotionEstimator_OTS::CreateMEData(const SearchParam& spar,const MEData*)
{
  MEData_OTS* medata = new MEData_OTS;
  medata->bcdata = d_blkcmp->CreateBCData(spar);
  medata->sparam = spar;

  return medata;
}

void MotionEstimator_OTS::SetSearchParams(MEData* medata, const Bitmap<Pixel>* bitmap,
                                           bool frame, bool top)
{
  MEData_OTS* med = dynamic_cast<MEData_OTS*>(medata);
  assert(med);

  c_medata = med;
  c_bcdata = med->bcdata;
  c_frame_flag = frame;
  c_top_flag   = top;
  c_img_width  = bitmap->AskWidth();
  c_img_height = bitmap->AskHeight();
  c_pixels_per_block = med->sparam.blk_width * med->sparam.blk_height;

  d_blkcmp->SetImageBitmap(*bitmap,frame,top);
}


bool MotionEstimator_OTS::Estimate(int x,int y, int& mx,int& my,
					   double& error,double upper_error_bound)
{
  MV_COMMON_PART
    /*
    cout << "OTS x0: " << x0 << " y0: " << y0 << endl;
  cout << "left: " << left << " right: " << right << " top: " << top << " bottom: " << bottom << endl;
    */

  uint32 maxerror = (uint32)(upper_error_bound*c_pixels_per_block);
  uint32 err;

  int stepsize = (c_medata->sparam.halfpel ? 2 : 1);

  uint32 lefterror,righterror,centererror;
  int hdirection;
  cmppos.ref_y = y0;
  cmppos.ref_x = x0;   d_blkcmp->Error(c_bcdata,cmppos,centererror,MASK31);
  cmppos.ref_x = x0-stepsize; if (x0-stepsize>=left)  d_blkcmp->Error(c_bcdata,cmppos,lefterror  ,MASK31);
                              else lefterror=MASK31;
  cmppos.ref_x = x0+stepsize; if (x0+stepsize<=right) d_blkcmp->Error(c_bcdata,cmppos,righterror ,MASK31);
                              else righterror=MASK31;

  if (centererror <= lefterror && centererror <= righterror)
    {
      mx=my=0;
      error=((double)centererror)/c_pixels_per_block;
      //    cout << "OTS CENTER\n";
      return true;
    }

  if (lefterror<righterror)  hdirection=-stepsize; else hdirection=stepsize;

  uint32 minerror=min(righterror,lefterror);
  int motx=2*hdirection;

  //if (hdirection>0) cout << "RIGHT\n"; else cout << "LEFT\n";

  while (motx>=left && motx<=right)
    {
      uint32 err;
      cmppos.ref_x = x0+motx;
      d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);

      if (err<minerror)
        { minerror=err; motx += hdirection; } // cout << "H-STEP\n"; }
      else
        { motx -= hdirection; break; }
    }

  cmppos.ref_x = x0+motx;
  mx = motx;


  uint32 uperror,downerror;
  cmppos.ref_y = y0-stepsize; if (y0-stepsize>=top)    d_blkcmp->Error(c_bcdata,cmppos,uperror  ,MASK31);
                              else uperror=MASK31;
  cmppos.ref_y = y0+stepsize; if (y0+stepsize<=bottom) d_blkcmp->Error(c_bcdata,cmppos,downerror,MASK31);
                              else downerror=MASK31;

  if (minerror <= uperror && minerror <= downerror)
    {
      //      cout << "HOR\n";
      my=0;
      error=((double)minerror)/c_pixels_per_block;
      return true;
    }

  int vdirection;
  if (uperror<downerror)  vdirection=-stepsize; else vdirection=stepsize;

  minerror=min(uperror,downerror);
  int moty=2*vdirection;

  //if (vdirection>0) cout << "DOWN\n"; else cout << "UP\n";

  while (moty>=top && moty<=bottom)
    {
      uint32 err;
      cmppos.ref_y = y0+moty;
      d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);

      if (err<minerror)
        { minerror=err; moty += vdirection; }// cout << "V-STEP\n"; }
      else
        { moty -= vdirection; break; }
    }

  my = moty;



  // Half-Pels drumrum testen

  if (c_medata->sparam.halfpel &&
      motx-1 >= left && motx+1 <= right && moty-1 >= top && moty+1 <= bottom)
    {
      uint32 err;
      cmppos.ref_x=x0+motx-1; cmppos.ref_y=y0+moty-1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx-1; my=moty-1; }
      cmppos.ref_x=x0+motx  ; cmppos.ref_y=y0+moty-1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx  ; my=moty-1; }
      cmppos.ref_x=x0+motx+1; cmppos.ref_y=y0+moty-1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx+1; my=moty-1; }
      cmppos.ref_x=x0+motx-1; cmppos.ref_y=y0+moty  ; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx-1; my=moty  ; }
      cmppos.ref_x=x0+motx+1; cmppos.ref_y=y0+moty  ; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx+1; my=moty  ; }
      cmppos.ref_x=x0+motx-1; cmppos.ref_y=y0+moty+1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx-1; my=moty+1; }
      cmppos.ref_x=x0+motx  ; cmppos.ref_y=y0+moty+1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx  ; my=moty+1; }
      cmppos.ref_x=x0+motx+1; cmppos.ref_y=y0+moty+1; d_blkcmp->Error(c_bcdata,cmppos,err,MASK31);
      if (err<minerror) { minerror=err; mx=motx+1; my=moty+1; }
    }

  //  cout << "MV: " << mx << " / " << my << endl;

  return true;
}
