/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/ratecontrol.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "statistics/bitsestim.hh"
#include "options.hh"
#include <stdlib.h>

static double lastfact_I = 1.0;
static double lastfact_P = 1.0;
static double lastfact_B = 1.0;

static int32 bitpool=0;  // Bits, die noch zusaetzlich vergeben werden koennen.

const int nInterations=9;

#if 1
const double Fixed_PI_Ratio = 1.5;           // VERY-High-Complexity
const double Fixed_BI_Ratio = 3.0;

//const double Fixed_PI_Ratio = 2.0; // 1.5           // High-Complexity
//const double Fixed_BI_Ratio = 6.0; // 3.0
#else
const double Fixed_PI_Ratio =  4.0; // 1.5          // Low-Complexity
const double Fixed_BI_Ratio = 10.0; // 3.0
#endif

const int IOverhead = 1500;
const int POverhead =  900;
const int BOverhead =  300;


void Task_RateControl::Execute(ProcessData& data)
{
  if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
    {
      ShowTaskInformation("rate control",data.threadnr);
      cerr << "frame " << m_startframenr << " to " << m_endframenr << endl;
      EndShowTaskInformation();
    }

  AssignFrameSizes(m_startframenr,m_endframenr);
}


void AssignFrameSizes(int from,int to)
{
  int32 nImages    = to-from+1;
  int32 TotalBytes = options.Bitrate/8*nImages /25;     // TODO: 25 durch echte Framerate ersetzen

  assert(nImages>=1);
  //cout << "ASSIGNFRAMESIZES " << options.Bitrate << " " << nImages << " -> ToBytes " << TotalBytes << endl;

  if (bitpool/8+TotalBytes >= 300*nImages)
    {
      TotalBytes += bitpool/8; // Buffer-Mismatch aus vorherigen GOPs aufaddieren.
      bitpool = bitpool%8;
    }
  else
    {
      assert(bitpool<0);

      int32 NewBytes = TotalBytes/2;    // Nur die Haelfte der Bytes vergeben.

      int32 diff = TotalBytes-NewBytes; // andere Haelfte spaeter vergeben.
      bitpool += diff*8;
    }

  uint32 Np=0,Nb=0,Ni=0;



  // Anzahl der I-, P- und B-Frames zaehlen

  {for (int n=from;n<=to;n++)
    {
      ImageData* img = ImageData::GetNr(n);
      assert(img);

           if (img->pichdr.m_picture_coding_type == PICTYPE_B) Nb++;
      else if (img->pichdr.m_picture_coding_type == PICTYPE_P) Np++;
      else Ni++;
    }}



  // Bits einer GOP an die Frames verteilen

  uint32 ISize = (int)(TotalBytes/(Ni + Np/Fixed_PI_Ratio + Nb/Fixed_BI_Ratio));
  uint32 PSize = (int)(ISize/Fixed_PI_Ratio);
  uint32 BSize = (int)(ISize/Fixed_BI_Ratio);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
    {
      cout << "TotalBytes " << TotalBytes << endl;
      cout << "Ni: " << Ni << " Np: " << Np << " Nb: " << Nb << endl;
      cout << "ISize: " << ISize << " PSize: " << PSize << " BSize: " << BSize << "  in BYTES" << endl;
    }

  {for (int n=from;n<=to;n++)
    {
      ImageData* img = ImageData::GetNr(n);
           if (img->pichdr.m_picture_coding_type == PICTYPE_B) img->desired_size = BSize;
      else if (img->pichdr.m_picture_coding_type == PICTYPE_P) img->desired_size = PSize;
      else img->desired_size = ISize;
    }}
}


void AdjustFramesizePrediction(const ImageData* img)
{
  double fact = ((double)img->output.d_bufferidx)/((double)img->expected_size);

  /*
  cout << "ADJ BUFFER:   " << img->output.d_bufferidx << endl;
  cout << "ADJ EXPECTED: " << img->expected_size << endl;

  cout << "ADJ SETFACT: " << fact << endl;
  */

  double lastfact = 0; /* Compiler beruhigen */
  switch (img->pichdr.m_picture_coding_type)
    {
    case PICTYPE_I: lastfact=lastfact_I; break;
    case PICTYPE_P: lastfact=lastfact_P; break;
    case PICTYPE_B: lastfact=lastfact_B; break;
    }

  fact = (lastfact+fact)/2.0;
  if (fact>2.0) fact=2.0;
  if (fact<0.5) fact=0.5;

  switch (img->pichdr.m_picture_coding_type)
    {
    case PICTYPE_I: lastfact_I=fact; break;
    case PICTYPE_P: lastfact_P=fact; break;
    case PICTYPE_B: lastfact_B=fact; break;
    }
}

void SetQOffsetToMeetFrameSize(ImageData* img)
{
  double low_q  = -50.0;
  double high_q =  50.0;
  double q      =   0.0;        // (Anfangswert des) QScale-Offsets


  double fact; // Korrektorfaktor fuer Bits-Estimation.
  int32 DesiredSize = img->desired_size;

       if (img->pichdr.m_picture_coding_type == PICTYPE_B) { fact=lastfact_B; DesiredSize -= BOverhead; }
  else if (img->pichdr.m_picture_coding_type == PICTYPE_P) { fact=lastfact_P; DesiredSize -= POverhead; }
  else { fact=lastfact_I; DesiredSize -= IOverhead; }

  if (DesiredSize<0) DesiredSize = img->desired_size;    // Fuer den Notfall


  // Passenden Offset fuer 'q' finden.

   int32 FrameSize=0;
  double bestq=q;
  uint32 bestdiff=MASK31;

  {for (int i=0;i<nInterations;i++)
    {
      // cout << "FACT: " << fact << endl;
      FrameSize = (int)(EstimateBitsToCode(img,q) * fact/8);

      /*
      cout << "iteration: " << i << " q: " << q << " size: " << FrameSize
           << "  desired: " << img->desired_size << " (" << DesiredSize << ")" << endl;
      */

      if (((uint32)abs(DesiredSize - FrameSize))<bestdiff)
        {
          bestdiff=abs(DesiredSize - FrameSize);
          bestq = q;
        }

      if (FrameSize<DesiredSize) { high_q = q; q = (low_q +q)/2; }
      else                       { low_q  = q; q = (high_q+q)/2; }
    }}

  img->expected_size = FrameSize;
  // img->desired_size = FrameSize;


  // Offset auf alle MBs addieren.

  uint32 w = img->mblks.AskWidth();
  uint32 h = img->mblks.AskHeight();

  {for (uint32 y=0;y<h;y++)
    for (uint32 x=0;x<w;x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);
        mb.m_quantiser_scale += bestq;
      }}
}

void AddBufferMismatch(int32 bits)
{
  bitpool += bits;

  if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
    cout << "Add Mismatch: " << bits/8 << " bytes\n";
}
