/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "control/mehelpert.hh"
#include "coding/motionest_full.hh"
#include "coding/motionest_log.hh"
#include "math/dct2_arai.hh"
#include "mpeg/constants.hh"
#include "mpeg/utility.hh"
#include "mpeg/defaultmatrices.hh"
#include "image/image.hh"
#include "image/colorspaces.hh"
#include "visualize/writeppm.hh"
#include "global.hh"
#include "measure/rms.hh"

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <stdio.h>


MotionEstimationHelper::MotionEstimationHelper()
  : d_motionalgo(MEAlgo_Undefined),
    forward_motionest(NULL),
    backward_motionest(NULL)
{
  SetSearchRange(8,8,8,8);
  SetMotionEstAlgo(LogSearch);


  // Speicherbereiche fuer Makroblockpraediktoren vorbereiten

  Image<DiffPixel> mb_pred; mb_pred.CreateImage(YCrCb,16,16);

  DiffPixel** mb_pred_y =mb_pred.AskBitmap(Bitmap_Y ).AskFrame();
  DiffPixel** mb_pred_cb=mb_pred.AskBitmap(Bitmap_Cb).AskFrame();
  DiffPixel** mb_pred_cr=mb_pred.AskBitmap(Bitmap_Cr).AskFrame();

  Image<Pixel> mb_predcomp; mb_predcomp.CreateImage(YCrCb,16,16);

  Pixel** mb_predcomp_y =mb_predcomp.AskBitmap(Bitmap_Y ).AskFrame();
  Pixel** mb_predcomp_cb=mb_predcomp.AskBitmap(Bitmap_Cb).AskFrame();
  Pixel** mb_predcomp_cr=mb_predcomp.AskBitmap(Bitmap_Cr).AskFrame();

  // Speicherbereiche fuer die bidirektionale Praediktion bei B-Frames

  Image<DiffPixel> mb_bipred; mb_bipred.CreateImage(YCrCb,16,16);

  DiffPixel** mb_bipred_y =mb_bipred.AskBitmap(Bitmap_Y ).AskFrame();
  DiffPixel** mb_bipred_cb=mb_bipred.AskBitmap(Bitmap_Cb).AskFrame();
  DiffPixel** mb_bipred_cr=mb_bipred.AskBitmap(Bitmap_Cr).AskFrame();

  Image<Pixel> mb_bipredcomp; mb_bipredcomp.CreateImage(YCrCb,16,16);

  Pixel** mb_bipredcomp_y =mb_bipredcomp.AskBitmap(Bitmap_Y ).AskFrame();
  Pixel** mb_bipredcomp_cb=mb_bipredcomp.AskBitmap(Bitmap_Cb).AskFrame();
  Pixel** mb_bipredcomp_cr=mb_bipredcomp.AskBitmap(Bitmap_Cr).AskFrame();
}

MotionEstimatorHelper::~MotionEstimatorHelper()
{
  if (forward_motionest)  delete forward_motionest;
  if (backward_motionest) delete backward_motionest;
}

void MotionEstimatorHelper::SetMotionEstAlgo(enum MotionEstimationAlgo algo)
{
  if (algo == d_motionalgo)
    return;

  if (forward_motionest)  delete forward_motionest;
  if (backward_motionest) delete backward_motionest;

  switch (algo)
    {
    case FullSearch:
      forward_motionest  = new MotionEstimator_Full;
      backward_motionest = new MotionEstimator_Full;
      break;
    case LogSearch:
      forward_motionest  = new MotionEstimator_Logarithmic;
      backward_motionest = new MotionEstimator_Logarithmic;
      break;
    default:
      assert(0);
      break;
    }

  d_motionalgo = algo;
}


void MotionEstimatorHelper::SetSearchParam(const SearchParam* fwd,const SearchParam* bkw)
{
  d_sparam_fwd = *fwd;
  d_sparam_bkw = *bkw;
}


void MotionEstimatorHelper::SetReferenceBitmaps(const Bitmap<Pixel>* fwd,
                                                const Bitmap<Pixel>* bkw,
                                                const Bitmap<Pixel>* fwd2,
                                                const Bitmap<Pixel>* bkw2)
{
  d_sparam_fwd.bitmap = fwd;
  d_sparam_bkw.bitmap = bkw;

  assert(fwd2==NULL);  // TODO: Interlaced
  assert(bkw2==NULL);
}


  // Vorinitialisierungen fuer MB-Codierungsentscheidung

  Macroblock mb_inter;
  Macroblock mb_inter_bi;
  Macroblock mb_intra;

  mb_inter.m_IsIntra=false;
  mb_inter.m_HasMotionForward = true;

  mb_inter_bi.m_IsIntra=false;
  mb_inter_bi.m_HasMotionForward  = true;
  mb_inter_bi.m_HasMotionBackward = true;

  mb_intra.m_IsIntra=true;
  mb_intra.m_HasMotionForward = false;


  // --- Codierungsschleife ---

  do
    {
      if (d_pic_nr > 0) imgtocode = d_preproc.GetNextImage();


      // --- Bildcodierungs-typ bestimmen ---

      bool code_I = false;
      bool code_P = false;   // nur beachten, falls 'code_I == false'

      if ((d_pic_nr % d_gop_len)==0)
	{
	  d_syntax.WriteGOPHeader(d_gop_data);
	  code_I = true;
	}

      if ((d_pic_nr % d_p_distance)==0)
	code_P = true;

      // Letztes Bild im Stream P codieren, falls es eigentlich B-codiert werden sollte.
      if (!d_preproc.HasMoreInput() && !code_P && !code_I)
	code_P = true;



      /*  BILD CODIEREN */

      globalstats.SetCurrentFrameNr(d_pic_nr);

           if (code_I) globalstats.SetCodingType(PICTYPE_I);
      else if (code_P) globalstats.SetCodingType(PICTYPE_P);


      cerr << "\nencoding frame " << d_pic_nr << " as ";
      if (code_I) cerr << 'I'; else if (code_P) cerr << 'P'; else cerr << 'B';
      cerr << "-Frame ... "; cerr.flush();


      /* Hinweis: Hier wird zwar auf die Rot-Gruen-Blau-Bitmaps zugegriffen, wenn das Bild
         jedoch aus YUV-Bitmaps besteht, werden diese geliefert. */
      imgtocode->AskBitmap(Bitmap_Red  ).DuplicateBorder();
      imgtocode->AskBitmap(Bitmap_Green).DuplicateBorder();
      imgtocode->AskBitmap(Bitmap_Blue ).DuplicateBorder();

      switch (imgtocode->AskColorspace())
	{
	case RGB_Undefined:
	  cerr << "Colorspace transformation ... "; cerr.flush();
	  TransformRGB2YCrCb(*imgtocode,sh.m_chroma_format,d_encopts.m_GenerateMPEG1);
	  break;
	case YCrCb:
	  cerr << "Color subsampling ... "; cerr.flush();
	  SubsampleYCrCb(*imgtocode,sh.m_chroma_format,d_encopts.m_GenerateMPEG1);
	  break;
	default:
	  assert(0);
	  break;
	}


#if 0
      {
        const Bitmap<Pixel>& bm = imgtocode->AskBitmap_const(Bitmap_Y);
        const Pixel*const* pix = bm.AskFrame_const();
        const Pixel*const* pixu = imgtocode->AskBitmap_const(Bitmap_U).AskFrame_const();
        const Pixel*const* pixv = imgtocode->AskBitmap_const(Bitmap_V).AskFrame_const();

        long long v=0;
        for (int y=0;y<codedimg_PB.AskHeight();y++)
          for (int x=0;x<codedimg_PB.AskWidth();x++)
            {
              v += Varianz(pix,x*16,y*16);
            }

        long long meany=0,meanu=0,meanv=0;
        for (int y=0;y<bm.AskHeight();y++)
          for (int x=0;x<bm.AskWidth();x++)
            {
              meany += pix [y][x];
              meanu += pixu[y][x];
              meanv += pixv[y][x];
            }

        long long lastvar,lasty,lastu,lastv;
        cout << d_pic_nr << " " << v << " " << meany << " " << meanu << " " << meanv << " ";

        if (d_pic_nr==0) cout << "0 0 0 0\n";
        else
          {
            cout << abs(v-lastvar) << " "
                 << abs(meany-lasty) << " "
                 << abs(meanu-lastu) << " "
                 << abs(meanv-lastv) << endl;
          }
        lastvar=v; lasty=meany; lastu=meanu; lastv=meanv;
      }
#endif

      if (code_I)
	{
	  codedimg_I.m_pichdr.m_temporal_reference = d_pic_nr%d_gop_len;

	  //FillQScale(*imgtocode,codedimg_I);

	  cerr << "DCT ... "; cerr.flush();

	  dctcoder_pix.CodeDCTBlocks(*imgtocode,codedimg_I);
	  FinishPicture(codedimg_I,d_pic_nr,d_gop_len != 1);
          globalstats.SetPredictionCnt(nTotalMBs,0,0,0);
          if (d_show_rms) globalstats.SetRMS(CalcRMS(*imgtocode,*decoder.m_current_img));
          globalstats.FinishFrame();


#if 0
    {
      char buf[500];
      static int nr=0;
      sprintf(buf,"qwe%03d.ppm",nr++);
      ofstream ostr(buf);
      Image<Pixel>* vimg = CreateVisualizationImage(*imgtocode);
      WriteImageAsPPM(*vimg,ostr);
      delete vimg;
    }
#endif


	  delete imgtocode;
	}
      else if (code_P)
	{
	  cerr << "MotionVec ...\n"; cerr.flush();

	  codedimg_PB.m_pichdr.m_picture_coding_type = PICTYPE_P;
	  codedimg_PB.m_pichdr.m_temporal_reference  = d_pic_nr%d_gop_len;

	  //FillQScale(*imgtocode,codedimg_I);

	  Pixel** pix_y  = imgtocode->AskBitmap(Bitmap_Y ).AskFrame();
	  Pixel** pix_cr = imgtocode->AskBitmap(Bitmap_Cr).AskFrame();
	  Pixel** pix_cb = imgtocode->AskBitmap(Bitmap_Cb).AskFrame();

	  const Pixel*const* past_y  = decoder.m_future_img->AskBitmap_const(Bitmap_Y ).AskFrame_const();
	  const Pixel*const* past_cb = decoder.m_future_img->AskBitmap_const(Bitmap_Cb).AskFrame_const();
	  const Pixel*const* past_cr = decoder.m_future_img->AskBitmap_const(Bitmap_Cr).AskFrame_const();

  if (0)
    {
      static nr = 0;
      char buf[500];
      sprintf(buf,"vis%03d.ppm",nr++);
      ofstream ostr(buf);
      Image<Pixel>* vimg = CreateVisualizationImage(*decoder.m_future_img);
      WriteImageAsPPM(*vimg,ostr);
      delete vimg;
    }

          sparam_fwd.bitmap = &decoder.m_future_img->AskBitmap(Bitmap_Y);
          MEData* medata = forward_motionest->CreateMEData(sparam_fwd);
          forward_motionest->SetSearchParams(medata, &imgtocode->AskBitmap(Bitmap_Y),true);

	  int icnt=0,pcnt=0;

	  int mv_fwdpred_x=0,mv_fwdpred_y=0;
	  int mv_bkwpred_x=0,mv_bkwpred_y=0;

	  // dctcoder.SetPicHeader(codedimg_PB.m_pichdr);

	  for (int y=0;y<codedimg_PB.AskHeight();y++,(cerr << '.' << y), cerr.flush())
	    for (int x=0;x<codedimg_PB.AskWidth();x++)
	      {
		Macroblock& mb = codedimg_PB.AccessMB(y,x);
		mb.m_SkipCoding=false;

		mb.m_vector[0][0].m_habs = mv_fwdpred_x;
		mb.m_vector[0][0].m_vabs = mv_fwdpred_y;
		mb.m_vector[0][1].m_habs = mv_bkwpred_x;
		mb.m_vector[0][1].m_vabs = mv_bkwpred_y;

		// Besten Motionvektor bestimmen

		double err;

		bool block_found = forward_motionest->Estimate(x*16,y*16,
							       mb.m_vector[0][0].m_habs,
							       mb.m_vector[0][0].m_vabs,
							       err);

		assert(block_found);

		/* Jetzt werden zwei Makrobloecke erstellt. Einmal wird der Motionvektor als
		   Praediktion benutzt, das andere Mal der Block ohne Praediktion DCT codiert.
		   Bei jedem Macroblock werden die Anzahl der Bits abgeschaetzt, die zur Codierung
		   notwendig sind und die Codierungsart mit den wenigeren Bits gewaehlt. */

		// Praediktion erstellen

		MotionCompensateSingleSet(mb_predcomp_y,0,0,x*16,y*16,16,16,past_y,
					  mb.m_forward1.m_habs,mb.m_forward1.m_vabs);
		MotionCompensateSingleSet(mb_predcomp_cb,0,0,x*8,y*8,8,8,past_cb,
					  mb.m_forward1.m_habs/2,
					  mb.m_forward1.m_vabs/2);
		MotionCompensateSingleSet(mb_predcomp_cr,0,0,x*8,y*8,8,8,past_cr,
					  mb.m_forward1.m_habs/2,
					  mb.m_forward1.m_vabs/2);

		// Differenzbild erstellen
		
		{for (int yy=0;yy<16;yy++)
		  for (int xx=0;xx<16;xx++)
		    mb_pred_y[yy][xx] = ((DiffPixel)pix_y[y*16+yy][x*16+xx] -
                                         (DiffPixel)mb_predcomp_y[yy][xx]);
		}
                
                {
		for (int yy=0;yy<8;yy++)
		  for (int xx=0;xx<8;xx++)
		    {
		      mb_pred_cb[yy][xx] = ((DiffPixel)pix_cb[y*8+yy][x*8+xx] -
                                            (DiffPixel)mb_predcomp_cb[yy][xx]);
		      mb_pred_cr[yy][xx] = ((DiffPixel)pix_cr[y*8+yy][x*8+xx] -
                                            (DiffPixel)mb_predcomp_cr[yy][xx]);
		    }
		}

		// --- Beide Makrobloecke ausfuellen und DCT anwenden ---

		mb_inter.m_quantiser_scale_code = d_qscale;
		mb_intra.m_quantiser_scale_code = d_qscale;

                /*
		dctcoder.MBlk_FDCT( mb_pred  ,0,0,mb_inter);
		dctcoder.MBlk_FDCT(*imgtocode,x,y,mb_intra);
                */

                if (d_DisablePredictionCorrection)
                  {
                    for (int a=0;a<8;a++)
                      for (int b=0;b<8;b++)
                        for (int c=0;c<12;c++)
                          mb_inter.m_blks[c].m_coeff[a][b]=0;
                  }

		// Bessere Makroblockcodierungsart auswaehlen und richtigen MB entsprechend ausfuellen

		const Macroblock* bestmb;

                int vi,vp;
                int bi,bp;
                bool right=false;

                vi = Varianz(pix_y,x*16,y*16);
                vp = Varianz(mb_pred_y,0,0);

                /*
                cout << d_pic_nr << " " << setw(2) << x << " " << setw(2) << y << " ";
                cout << "VIntra: " << setw(6) << vi << "  ";
                cout << "VInter: " << setw(6) << vp << "  ";
                */
                /*
		cout << "BIntra: " << setw(4)<<(bi=BitsToCodeCoeff(mb_intra,codedimg_PB.m_pichdr))<<" ";
		cout << "BInter: " << setw(4)<<(bp=BitsToCodeCoeff(mb_inter,codedimg_PB.m_pichdr)+6)<<" ";
                */

                // vi = (int)(vi*1.2);

                if (bi<bp && vi<vp)
                  right=true;
                if (bi>bp && vi>vp)
                  right=true;

                if (right) n_right++;
                else n_false++;

                /*
		if (BitsToCodeCoeff(mb_intra,codedimg_PB.m_pichdr) <=
		    BitsToCodeCoeff(mb_inter,codedimg_PB.m_pichdr)+6)
                */

                if (vi<vp)
		  {
                    //cout << "INTRA <<<<<<<<< ";
		    bestmb = &mb_intra;

		    mb.m_IsIntra = true;
		    mb.m_HasMotionForward = false;

		    mv_fwdpred_x = 0;
		    mv_fwdpred_y = 0;
		    mv_bkwpred_x = 0;
		    mv_bkwpred_y = 0;

		    icnt++;

		dctcoder_pix.MBlk_FDCTQuant(*imgtocode,x,y,mb_intra);
		  }
		else
		  {
                    //cout << "INTER           ";
		    bestmb = &mb_inter;

		    mb.m_IsIntra = false;
		    mb.m_HasMotionForward = true;
		    mb.m_FieldDCT = false;
		    mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);

		    mv_fwdpred_x = mb.m_vector[0][0].m_habs;
		    mv_fwdpred_y = mb.m_vector[0][0].m_vabs;
		    mv_bkwpred_x = mb.m_vector[0][1].m_habs;
		    mv_bkwpred_y = mb.m_vector[0][1].m_vabs;

		    pcnt++;

		dctcoder_dpix.MBlk_FDCTQuant( mb_pred  ,0,0,mb_inter);
		  }

                //if (right) cout << "OK\n"; else cout << "nok\n";
		
		mb.m_HasMotionBackward = false;
		mb.m_quantiser_scale_code = d_qscale;
		
		for (int i=0;i<12;i++)
		  {
		    if (mb.m_CodeBlk[i]) mb.m_blks[i] = bestmb->m_blks[i];
		  }

		mb.m_SkipCoding = true;  // Haben wir ja schon codiert und die Daten reinkopiert.
	      }

	  cerr << endl
	       << icnt << " Intra Blocks   "
	       << pcnt << " Predictive Blocks\n";

          delete medata;
	  FinishPicture(codedimg_PB,d_pic_nr,true);
          globalstats.SetPredictionCnt(icnt,pcnt,0,0);
          if (d_show_rms) globalstats.SetRMS(CalcRMS(*imgtocode,*decoder.m_current_img));
          globalstats.FinishFrame();
	  delete imgtocode;
	}
      else // B-Pictures
	{
	  cerr << " coding deferred\n";
	  b_frame_queue.Append(imgtocode);
	}


      // Alle noch anstehenden B-Frames codieren, falls das letzte ein I- oder P- Frame war.

      MEData* medata_fwd = 0;
      MEData* medata_bkw = 0;

      if ((code_I || code_P) && !b_frame_queue.IsEmpty())
	{
          sparam_fwd.bitmap = &decoder.m_past_img->AskBitmap(Bitmap_Y);
          medata_fwd = forward_motionest->CreateMEData(sparam_fwd);

          sparam_bkw.bitmap = &decoder.m_future_img->AskBitmap(Bitmap_Y);
          medata_bkw = backward_motionest->CreateMEData(sparam_bkw);
	}

      while ((code_I || code_P) && !b_frame_queue.IsEmpty())
	{
	  codedimg_PB.m_pichdr.m_picture_coding_type = PICTYPE_B;
	  codedimg_PB.m_pichdr.m_temporal_reference  = (d_pic_nr-b_frame_queue.Length())%d_gop_len;

	  cerr << "\nCoding deferred B-frame (" << d_pic_nr-b_frame_queue.Length() << ")\n";

	  Image<Pixel>* Bframe = b_frame_queue.AskHead();

	  //FillQScale(*Bframe,codedimg_I);
	  
	  cerr << "bidirectional MotionVec ...\n"; cerr.flush();

	  Pixel** pix_y  = Bframe->AskBitmap(Bitmap_Y ).AskFrame();
	  Pixel** pix_cr = Bframe->AskBitmap(Bitmap_Cr).AskFrame();
	  Pixel** pix_cb = Bframe->AskBitmap(Bitmap_Cb).AskFrame();

	  const Pixel*const* past_y  = decoder.m_past_img->AskBitmap_const(Bitmap_Y ).AskFrame_const();
	  const Pixel*const* past_cb = decoder.m_past_img->AskBitmap_const(Bitmap_Cb).AskFrame_const();
	  const Pixel*const* past_cr = decoder.m_past_img->AskBitmap_const(Bitmap_Cr).AskFrame_const();

	  const Pixel*const* future_y  = decoder.m_future_img->AskBitmap_const(Bitmap_Y ).AskFrame_const();
	  const Pixel*const* future_cb = decoder.m_future_img->AskBitmap_const(Bitmap_Cb).AskFrame_const();
	  const Pixel*const* future_cr = decoder.m_future_img->AskBitmap_const(Bitmap_Cr).AskFrame_const();


          forward_motionest ->SetSearchParams(medata_fwd, &Bframe->AskBitmap(Bitmap_Y),true);
          backward_motionest->SetSearchParams(medata_bkw, &Bframe->AskBitmap(Bitmap_Y),true);

	  int icnt=0,pcnt=0,fcnt=0,bcnt=0,dcnt=0;

	  int mv_pred_x=0,mv_pred_y=0;

	  // dctcoder.SetPicHeader(codedimg_PB.m_pichdr);

	  for (int y=0;y<codedimg_PB.AskHeight();y++,(cerr << '.' << y), cerr.flush())
	    for (int x=0;x<codedimg_PB.AskWidth();x++)
	      {
		Macroblock& mb = codedimg_PB.AccessMB(y,x);
		mb.m_SkipCoding=false;

		mb.m_vector[0][0].m_habs = mv_pred_x;
		mb.m_vector[0][0].m_vabs = mv_pred_y;

		// Besten Motionvektor bestimmen

		double err_forward,err_backward;

		bool forward_block_found, backward_block_found;
		forward_block_found = forward_motionest->Estimate(x*16,y*16,
								  mb.m_vector[0][0].m_habs,
								  mb.m_vector[0][0].m_vabs,
								  err_forward);
		backward_block_found = backward_motionest->Estimate(x*16,y*16,
								    mb.m_vector[0][1].m_habs,
								    mb.m_vector[0][1].m_vabs,
								    err_backward);
		
		assert(forward_block_found);
		assert(backward_block_found);

		bool UseForwardPred = (err_forward<err_backward);


		/* Jetzt werden zwei Makrobloecke erstellt. Einmal wird der Motionvektor als
		   Praediktion benutzt, das andere Mal der Block ohne Praediktion DCT codiert.
		   Bei jedem Macroblock werden die Anzahl der Bits abgeschaetzt, die zur Codierung
		   notwendig sind und die Codierungsart mit den wenigeren Bits gewaehlt. */

		// Praediktion erstellen

		if (UseForwardPred)
		  {
		    MotionCompensateSingleSet(mb_predcomp_y,0,0,x*16,y*16,16,16,past_y,
					      mb.m_forward1.m_habs,mb.m_forward1.m_vabs);
		    MotionCompensateSingleSet(mb_predcomp_cb,0,0,x*8,y*8,8,8,past_cb,
					      mb.m_forward1.m_habs/2,
					      mb.m_forward1.m_vabs/2);
		    MotionCompensateSingleSet(mb_predcomp_cr,0,0,x*8,y*8,8,8,past_cr,
					      mb.m_forward1.m_habs/2,
					      mb.m_forward1.m_vabs/2);
		  }
		else
		  {
		    MotionCompensateSingleSet(mb_predcomp_y,0,0,x*16,y*16,16,16,future_y,
					      mb.m_backward1.m_habs,mb.m_backward1.m_vabs);
		    MotionCompensateSingleSet(mb_predcomp_cb,0,0,x*8,y*8,8,8,future_cb,
					      mb.m_backward1.m_habs/2,
					      mb.m_backward1.m_vabs/2);
		    MotionCompensateSingleSet(mb_predcomp_cr,0,0,x*8,y*8,8,8,future_cr,
					      mb.m_backward1.m_habs/2,
					      mb.m_backward1.m_vabs/2);
		  }


		// Und jetzt nochmals mit beiden MVs

                MotionCompensateDoubleSet(mb_bipredcomp_y,0,0,x*16,y*16,16,16,
					  past_y,  mb.m_forward1 .m_habs,mb.m_forward1 .m_vabs,
					  future_y,mb.m_backward1.m_habs,mb.m_backward1.m_vabs);
		MotionCompensateDoubleSet(mb_bipredcomp_cb,0,0,x*8,y*8,8,8,
					  past_cb  ,mb.m_forward1 .m_habs/2,mb.m_forward1 .m_vabs/2,
					  future_cb,mb.m_backward1.m_habs/2,mb.m_backward1.m_vabs/2);
		MotionCompensateDoubleSet(mb_bipredcomp_cr,0,0,x*8,y*8,8,8,
					  past_cr  ,mb.m_forward1 .m_habs/2,mb.m_forward1 .m_vabs/2,
					  future_cr,mb.m_backward1.m_habs/2,mb.m_backward1.m_vabs/2);


		// Differenzbilder erstellen
		
		{for (int yy=0;yy<16;yy++)
		  for (int xx=0;xx<16;xx++)
		    {
		      mb_pred_y  [yy][xx] = ((DiffPixel)pix_y[y*16+yy][x*16+xx] - 
                                             (DiffPixel)mb_predcomp_y  [yy][xx]);
		      mb_bipred_y[yy][xx] = ((DiffPixel)pix_y[y*16+yy][x*16+xx] -
                                             (DiffPixel)mb_bipredcomp_y[yy][xx]);
		    }
                }

		{for (int yy=0;yy<8;yy++)
		  for (int xx=0;xx<8;xx++)
		    {
		      mb_pred_cb  [yy][xx] = ((DiffPixel)pix_cb[y*8+yy][x*8+xx] -
                                              (DiffPixel)mb_predcomp_cb  [yy][xx]);
		      mb_pred_cr  [yy][xx] = ((DiffPixel)pix_cr[y*8+yy][x*8+xx] -
                                              (DiffPixel)mb_predcomp_cr  [yy][xx]);
		      mb_bipred_cb[yy][xx] = ((DiffPixel)pix_cb[y*8+yy][x*8+xx] -
                                              (DiffPixel)mb_bipredcomp_cb[yy][xx]);
		      mb_bipred_cr[yy][xx] = ((DiffPixel)pix_cr[y*8+yy][x*8+xx] -
                                              (DiffPixel)mb_bipredcomp_cr[yy][xx]);
		    }
		}

		// --- Beide Makrobloecke ausfuellen und DCT anwenden ---

		mb_inter   .m_quantiser_scale_code = d_qscale;
		mb_inter_bi.m_quantiser_scale_code = d_qscale;
		mb_intra   .m_quantiser_scale_code = d_qscale;
                /*
		dctcoder.MBlk_FDCT( mb_pred  ,0,0,mb_inter   );
		dctcoder.MBlk_FDCT( mb_bipred,0,0,mb_inter_bi);
		dctcoder.MBlk_FDCT(*Bframe   ,x,y,mb_intra   );
                */
                if (d_DisablePredictionCorrection)
                  {
                    for (int a=0;a<8;a++)
                      for (int b=0;b<8;b++)
                        for (int c=0;c<12;c++)
                          {
                            mb_inter.m_blks[c].m_coeff[a][b]=0;
                            mb_inter_bi.m_blks[c].m_coeff[a][b]=0;
                          }
                  }

		// Bessere Makroblockcodierungsart auswaehlen und richtigen MB entsprechend ausfuellen

		const Macroblock* bestmb;
#if 0
		int inter_bits   = BitsToCodeCoeff(mb_inter   ,codedimg_PB.m_pichdr);
		int interbi_bits = BitsToCodeCoeff(mb_inter_bi,codedimg_PB.m_pichdr);
		int intra_bits   = BitsToCodeCoeff(mb_intra   ,codedimg_PB.m_pichdr);

#endif
                int inter_bits   = Varianz(mb_pred_y  ,0,0);
                int interbi_bits = Varianz(mb_bipred_y,0,0);
                int intra_bits   = Varianz(pix_y,x*16,y*16);

		int minbits = min(inter_bits,min(interbi_bits,intra_bits));

		if (inter_bits == minbits)
		  {
		    if (UseForwardPred) fcnt++; else bcnt++;

		    bestmb = &mb_inter;

		    mb.m_IsIntra = false;
		    mb.m_HasMotionForward  =  UseForwardPred;
		    mb.m_HasMotionBackward = !UseForwardPred;
		    mb.m_FieldDCT = false;
		    mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);

		    mv_pred_x = mb.m_vector[0][0].m_habs;
		    mv_pred_y = mb.m_vector[0][0].m_vabs;

		    pcnt++;

                    dctcoder_dpix.MBlk_FDCTQuant( mb_pred  ,0,0,mb_inter   );
		  }
		else if (intra_bits == minbits)
		  {
		    bestmb = &mb_intra;

		    mb.m_IsIntra = true;
		    mb.m_HasMotionForward  = false;
		    mb.m_HasMotionBackward = false;

		    mv_pred_x = 0;
		    mv_pred_y = 0;

		    icnt++;

                    dctcoder_pix.MBlk_FDCTQuant(*Bframe   ,x,y,mb_intra   );
		  }
		else
		  {
		    dcnt++;

		    bestmb = &mb_inter_bi;

		    mb.m_IsIntra = false;
		    mb.m_HasMotionForward  = true;
		    mb.m_HasMotionBackward = true;
		    mb.m_FieldDCT = false;
		    mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);

		    mv_pred_x = mb.m_vector[0][0].m_habs;
		    mv_pred_y = mb.m_vector[0][0].m_vabs;

		    pcnt++;

                    dctcoder_dpix.MBlk_FDCTQuant( mb_bipred,0,0,mb_inter_bi);
		  }

		mb.m_quantiser_scale_code = d_qscale;
		
		for (int i=0;i<12;i++)
		  {
		    if (mb.m_CodeBlk[i]) mb.m_blks[i] = bestmb->m_blks[i];
		  }

		mb.m_SkipCoding = true;  // Haben wir ja schon codiert und die Daten reinkopiert.
	      }

	  cerr << endl
	       << icnt << " Intra Blocks   "
	       << pcnt << " Predictive Blocks ("
	       << fcnt << " forward "
	       << bcnt << " backward "
	       << dcnt << " bidirectional)\n";

	  globalstats.SetCurrentFrameNr(d_pic_nr-b_frame_queue.Length());
          globalstats.SetCodingType(PICTYPE_B);

	  FinishPicture(codedimg_PB, d_pic_nr-b_frame_queue.Length() , false);
          globalstats.SetPredictionCnt(icnt,fcnt,bcnt,dcnt);
          if (d_show_rms) globalstats.SetRMS(CalcRMS(*Bframe,*decoder.m_current_img));
          globalstats.FinishFrame();

	  b_frame_queue.RemoveHead();
	  delete Bframe;
	}

      if (medata_fwd) delete medata_fwd;
      if (medata_bkw) delete medata_bkw;

      d_pic_nr++;
      d_gop_data.AdvanceOnePicture(d_preproc.AskFramerate100()/100);

    } while (d_preproc.HasMoreInput());
}

static uint8 offs2codetab[2*4096];
static uint8* offs2code = &offs2codetab[4096];

static class DummyInitClass
{
public:
  DummyInitClass()
    {
      int i=0;
      for (;i<  16;i++) offs2code[i]=1;
      for (;i<  32;i++) offs2code[i]=2;
      for (;i<  64;i++) offs2code[i]=3;
      for (;i< 128;i++) offs2code[i]=4;
      for (;i< 256;i++) offs2code[i]=5;
      for (;i< 512;i++) offs2code[i]=6;
      for (;i<1024;i++) offs2code[i]=7;
      for (;i<2048;i++) offs2code[i]=8;
      for (;i<4096;i++) offs2code[i]=9;

      i=0;
      for (;i>=  -16;i--) offs2code[i]=1;
      for (;i>=  -32;i--) offs2code[i]=2;
      for (;i>=  -64;i--) offs2code[i]=3;
      for (;i>= -128;i--) offs2code[i]=4;
      for (;i>= -256;i--) offs2code[i]=5;
      for (;i>= -512;i--) offs2code[i]=6;
      for (;i>=-1024;i--) offs2code[i]=7;
      for (;i>=-2048;i--) offs2code[i]=8;
      for (;i>=-4096;i--) offs2code[i]=9;
    }
} dummy;

void EncoderControl_Simple::AdjustMotionRange(CodedImage& codedimg)
{
  if (codedimg.m_pichdr.m_picture_coding_type == PICTYPE_I)
    return;

  int fwd_hrange[10],bkw_hrange[10];
  int fwd_vrange[10],bkw_vrange[10];
  int fwd_total=0,   bkw_total=0;

  {for (int i=1;i<=9;i++)
    fwd_hrange[i]=fwd_vrange[i]=bkw_hrange[i]=bkw_vrange[i]=0;}

  for (int y=0;y<codedimg.AskHeight();y++)
    for (int x=0;x<codedimg.AskWidth();x++)
      {
        Macroblock& mb = codedimg.AccessMB(y,x);
        if (mb.m_IsIntra)
          continue;

        if (mb.m_HasMotionForward)
          {
            fwd_total++;
            fwd_hrange[offs2code[mb.m_forward1.m_habs]]++;
            fwd_vrange[offs2code[mb.m_forward1.m_vabs]]++;
          }
        if (mb.m_HasMotionBackward)
          {
            bkw_total++;
            bkw_hrange[offs2code[mb.m_backward1.m_habs]]++;
            bkw_vrange[offs2code[mb.m_backward1.m_vabs]]++;
          }
      }

  if (fwd_total || bkw_total)
    {
      int maxfwdh=1,maxbkwh=1;
      int maxfwdv=1,maxbkwv=1;
      for (int i=9;i>=1;i--)
        {
          if (fwd_hrange[i] && maxfwdh<i) maxfwdh=i;
          if (fwd_vrange[i] && maxfwdv<i) maxfwdv=i;
          if (bkw_hrange[i] && maxbkwh<i) maxbkwh=i;
          if (bkw_vrange[i] && maxbkwv<i) maxbkwv=i;
        }

      if (fwd_total || bkw_total) cerr << "\nMaximum MV-fcodes used:";
      if (fwd_total) cerr << "   fwd " << maxfwdh << ";" << maxfwdv;
      if (bkw_total) cerr << "   bkw " << maxbkwh << ";" << maxbkwv;
      if (fwd_total || bkw_total) cerr << endl;
      
      assert(0);
      codedimg.m_pichdr.m_f_code[0][0] = maxfwdh;
      codedimg.m_pichdr.m_f_code[0][1] = maxfwdv;
      codedimg.m_pichdr.m_f_code[1][0] = maxbkwh;
      codedimg.m_pichdr.m_f_code[1][1] = maxbkwv;
    }
}

void EncoderControl_Simple::FinishPicture(CodedImage& codedimg,int nr,bool decode)
{
  cerr << "Writing ... "; cerr.flush();

  AdjustMotionRange(codedimg);

  d_syntax.WritePicture(codedimg);

  if (decode || d_visualize || d_show_rms)
    {
      cerr << "Decode"; cerr.flush();
      decoder.Decode(codedimg);
    }

  cerr << endl;

  if (d_visualize)
    {
      bool CodType_I = (codedimg.m_pichdr.m_picture_coding_type == PICTYPE_I);

      char buf[500];
      sprintf(buf,d_visualizetemplate,nr);
      ofstream ostr(buf);
      Image<Pixel>* vimg = CreateVisualizationImage(*decoder.m_current_img);
      //FillBlockClass(*vimg,d_qscale_field);
      FillBlockTypes(*vimg,codedimg);
      if (!CodType_I) FillMotionVectors(*vimg,codedimg);
      WriteImageAsPPM(*vimg,ostr);
      delete vimg;
    }
}


void EncoderControl_Simple::FillSequenceHeader(SequenceHeaderData& sh)
{
  sh.m_width  = d_width;
  sh.m_height = d_height;
  sh.m_aspect_ratio_information = SAR_Square;
  sh.m_framerate100 = d_preproc.AskFramerate100();
  sh.m_bit_rate = MASK30;
  sh.m_vbv_buffer_size = MASK18;
  sh.m_constrained_parameters_flag = false;
  sh.m_load_intra_quantiser_matrix    = false;
  sh.m_load_nonintra_quantiser_matrix = false;

  sh.m_profile = PROFILE_Main;
  sh.m_level   = LEVEL_Main;
  sh.m_progressive_sequence = d_preproc.IsProgressiveSequence();
  sh.m_chroma_format = CHROMA_420;
  sh.m_low_delay = false;
}

void EncoderControl_Simple::FillPictureHeader(PictureHeaderData& pd)
{
  pd.m_temporal_reference=0;

  pd.m_vbv_delay = 0xFFFF;
  
  pd.m_full_pel_forward_vector  = false;
  pd.m_full_pel_backward_vector = false;

  assert(0);
  pd.m_f_code[0][0]=d_fcode[0][0];
  pd.m_f_code[0][1]=d_fcode[0][1];
  pd.m_f_code[1][0]=d_fcode[1][0];
  pd.m_f_code[1][1]=d_fcode[1][1];

  pd.m_intra_dc_precision= d_intra_dc_precision;
  pd.m_picture_structure = PICSTRUCT_FramePicture;

  pd.m_top_field_first = true;
  pd.m_frame_pred_frame_dct = true;
  pd.m_concealment_motion_vectors = false;
  pd.m_q_scale_type = d_qscale_type;
  pd.m_intra_vlc_format = d_intra_vlc_format;
  pd.m_alternate_scan = d_alternate_scan;
  pd.m_repeat_first_field = false;
  pd.m_progressive_frame = true;
  
  pd.m_composite_display_flag = false;
}

void EncoderControl_Simple::FillQScale(const Image<Pixel>& img,CodedImage& codedimg)
{
  d_qscale_dctcoder.CodeDCTBlocks(img,codedimg);

  for (int y=0;y<codedimg.AskHeight();y++)
    for (int x=0;x<codedimg.AskWidth();x++)
      {
	int nSmooth =0;
	int nEdge   =0;
	int nTexture=0;

	for (int blk=0;blk<4;blk++)
	  {
	    double Eh,Ev,Ed,Ea,Em,EM,EmM;
	    double T1 = 0.5;
	    double T2 = 0.3;
	    double T3 = 0.1;
	    const DCTBlock& dctblk = codedimg.AskMB(y,x).m_blks[blk];

#define Sq(x) ((x)*(x))

	    Ev = Sq(dctblk.m_coeff[2][0]) + Sq(dctblk.m_coeff[2][1]) +
	         Sq(dctblk.m_coeff[3][0]) + Sq(dctblk.m_coeff[3][1]) +
	         Sq(dctblk.m_coeff[4][0]) + Sq(dctblk.m_coeff[4][1]);
	    Eh = Sq(dctblk.m_coeff[0][2]) + Sq(dctblk.m_coeff[1][2]) +
	         Sq(dctblk.m_coeff[0][3]) + Sq(dctblk.m_coeff[1][3]) +
	         Sq(dctblk.m_coeff[0][4]) + Sq(dctblk.m_coeff[1][4]);
	    Ed = Sq(dctblk.m_coeff[2][2]) + Sq(dctblk.m_coeff[2][3]) +
	         Sq(dctblk.m_coeff[3][2]) + Sq(dctblk.m_coeff[3][3]);
	    Ea = (Ev+Eh+Ed)/3;
	    Em = min(Ev,min(Eh,Ed));
	    EM = max(Ev,max(Eh,Ed));
	    EmM = Em/EM;

	    // cout << Ea << "  " << Em << "  " << EmM << endl;

	    if (Ea<T1)
	      { nSmooth++; }
	    else if (Ea>=T1 && Em>T2 && EmM>T3)
	      { nTexture++; }
	    else
	      { nEdge++; }
	  }

	int maxi = max(nSmooth,max(nTexture,nEdge));
             if (maxi==nSmooth)  d_qscale_field[y][x]=0;
	else if (maxi==nTexture) d_qscale_field[y][x]=1;
	else if (maxi==nEdge)    d_qscale_field[y][x]=2;
      }
}

#include "utility/queue.cc"

