/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "math/dct1_arai.hh"
#include <math.h>
#include <assert.h>

const DCTFloat A1 = 0.707106781;
const DCTFloat A2 = 0.5411961;
const DCTFloat A3 = 0.707106781;
const DCTFloat A4 = 1.306562965;
const DCTFloat A5 = 0.382683432;

const DCTFloat A1INV = (1.0/A1);
const DCTFloat A2INV = (1.0/A2);
const DCTFloat A3INV = (1.0/A3);
const DCTFloat A4INV = (1.0/A4);
const DCTFloat A5INV = (1.0/A5);

const DCTFloat OutputScaleFactor[8]=
{
  0.35355339,  // 1/2*sqrt(2)
  0.254897789, // sec(PI*u/16)/4    u=1
  0.27059805,  //  ...              u=2
  0.300672443, //  ...              u=3
  0.35355339,
  0.449988111,
  0.653281482,
  1.281457724
};


DCT1_AraiAguiNakajima::DCT1_AraiAguiNakajima()
{
  RemoveOutputScalefactors();
}

void DCT1_AraiAguiNakajima::RemoveOutputScalefactors()
{
  DCTFloat f[8];
  for (int i=0;i<8;i++) f[i]=1.0;
  SetOutputScalefactors(f);
}

void DCT1_AraiAguiNakajima::SetOutputScalefactors(DCTFloat fact[8])
{
  for (int i=0;i<8;i++)
    {
      assert(fact[i] != 0.0);

      d_output_factors[i]     =      fact[i] * OutputScaleFactor[i];
      d_output_factors_INV[i] = 1.0/(fact[i] * OutputScaleFactor[i]);
    }
}


void DCT1_AraiAguiNakajima::PerformFDCT1(const DCTFloat in[8],DCTFloat out[8])
{
  DCTFloat tmp [8];
  DCTFloat tmp2[8];

  tmp[0] = in[0]+in[7];
  tmp[1] = in[1]+in[6];
  tmp[2] = in[2]+in[5];
  tmp[3] = in[3]+in[4];
  tmp[4] = in[3]-in[4];
  tmp[5] = in[2]-in[5];
  tmp[6] = in[1]-in[6];
  tmp[7] = in[0]-in[7];

  tmp2[0] =   tmp[0] + tmp[3];
  tmp2[1] =   tmp[1] + tmp[2];
  tmp2[2] =   tmp[1] - tmp[2];
  tmp2[3] =   tmp[0] - tmp[3];
  tmp2[4] = - tmp[4] - tmp[5];
  tmp2[5] =   tmp[5] + tmp[6];
  tmp2[6] =   tmp[6] + tmp[7];
  tmp2[7] =   tmp[7];

  tmp[0] = tmp2[0] + tmp2[1];
  tmp[1] = tmp2[0] - tmp2[1];
  tmp[2] = tmp2[2] + tmp2[3];
  tmp[3] = tmp2[3];
  tmp[4] = tmp2[4];
  tmp[5] = tmp2[5];
  tmp[6] = tmp2[6];
  tmp[7] = tmp2[7];

  DCTFloat a5 = (tmp2[4]+tmp2[6]) * A5;

  tmp2[0] =       tmp[0];
  tmp2[1] =       tmp[1];
  tmp2[2] =  A1 * tmp[2];
  tmp2[3] =       tmp[3];
  tmp2[4] = -A2 * tmp[4] - a5;
  tmp2[5] =  A3 * tmp[5];
  tmp2[6] =  A4 * tmp[6] - a5;
  tmp2[7] =       tmp[7];

  tmp[0] = tmp2[0];
  tmp[1] = tmp2[1];
  tmp[2] = tmp2[2] + tmp2[3];
  tmp[3] = tmp2[3] - tmp2[2];
  tmp[4] = tmp2[4];
  tmp[5] = tmp2[5] + tmp2[7];
  tmp[6] = tmp2[6];
  tmp[7] = tmp2[7] - tmp2[5];

  tmp2[0] = tmp[0];
  tmp2[1] = tmp[1];
  tmp2[2] = tmp[2];
  tmp2[3] = tmp[3];
  tmp2[4] = tmp[4] + tmp[7];
  tmp2[5] = tmp[5] + tmp[6];
  tmp2[6] = tmp[5] - tmp[6];
  tmp2[7] = tmp[7] - tmp[4];

  out[0] = tmp2[0];
  out[4] = tmp2[1];
  out[2] = tmp2[2];
  out[6] = tmp2[3];
  out[5] = tmp2[4];
  out[1] = tmp2[5];
  out[7] = tmp2[6];
  out[3] = tmp2[7];

  for (int i=0;i<8;i++)
    out[i] *= d_output_factors[i];
}


void DCT1_AraiAguiNakajima::PerformIDCT1(const DCTFloat in[8],DCTFloat out[8])
{
  // ------------------------------ FALSCH --------------------

  assert(0);

  DCTFloat tmp [8];
  DCTFloat tmp2[8];

  for (int i=0;i<8;i++)
    out[i] *= d_output_factors_INV[i];

  tmp[0] = in[0];
  tmp[1] = in[4];
  tmp[2] = in[2];
  tmp[3] = in[6];
  tmp[4] = in[5];
  tmp[5] = in[1];
  tmp[6] = in[7];
  tmp[7] = in[3];

  tmp2[0] = tmp[0];
  tmp2[1] = tmp[1];
  tmp2[2] = tmp[2];
  tmp2[3] = tmp[3];
  tmp2[4] = tmp[4] - tmp[7];
  tmp2[5] = tmp[5] + tmp[6];
  tmp2[6] = tmp[5] - tmp[6];
  tmp2[7] = tmp[7] + tmp[4];

  tmp[0] = tmp2[0];
  tmp[1] = tmp2[1];
  tmp[2] = tmp2[2] - tmp2[3];
  tmp[3] = tmp2[3] + tmp2[2];
  tmp[4] = tmp2[4];
  tmp[5] = tmp2[5] - tmp2[7];
  tmp[6] = tmp2[6];
  tmp[7] = tmp2[7] + tmp2[5];

  DCTFloat a5 = (tmp[6] + tmp[4])*(-A5INV);

  tmp2[0] = tmp[0];
  tmp2[1] = tmp[1];
  tmp2[2] = tmp[2] * A1INV;
  tmp2[3] = tmp[3];
  tmp2[4] = tmp[4]*(-A2INV) + a5;
  tmp2[5] = tmp[5]*A3INV;
  tmp2[6] = tmp[6]*A4INV    + a5;
  tmp2[7] = tmp[7];
  
  tmp[0] = tmp2[0] + tmp2[1];
  tmp[1] = tmp2[0] - tmp2[1];
  tmp[2] = tmp2[2];
  tmp[3] = tmp2[2] + tmp2[3];
  tmp[4] = tmp2[4];
  tmp[5] = tmp2[5];
  tmp[6] = tmp2[6];
  tmp[7] = tmp2[7];

  tmp2[0] =  tmp[0] + tmp[3];
  tmp2[1] =  tmp[1] + tmp[2];
  tmp2[2] =  tmp[1] - tmp[2];
  tmp2[3] =  tmp[0] - tmp[3];
  tmp2[4] = -tmp[4];
  tmp2[5] =  tmp[5] - tmp[4];
  tmp2[6] =  tmp[6] + tmp[5];
  tmp2[7] =  tmp[7] + tmp[6];

  out[0] = tmp2[0] + tmp2[7];
  out[1] = tmp2[1] + tmp2[6];
  out[2] = tmp2[2] + tmp2[5];
  out[3] = tmp2[3] + tmp2[4];
  out[4] = tmp2[3] - tmp2[4];
  out[5] = tmp2[2] - tmp2[5];
  out[6] = tmp2[1] - tmp2[6];
  out[7] = tmp2[0] - tmp2[7];
}
